#cython: language_level=3str


cdef class _CSCMatrix:
    """
    CSC matrix as specified by SPRAL (one-indexed, lower triangular part only)
    """
    cdef const int[:] rows
    cdef const int[:] cols
    cdef const double[:] data
    cdef object shape



#cython: language_level=3str

from enum import Enum
from typing import Optional

import numpy as np
import scipy as sp


cdef extern from "spral.h":
    struct spral_lsmr_options:
        double atol
        double btol
        double conlim
        int ctest
        int itnlim
        int itn_test
        int localSize
        int print_freq_head
        int print_freq_itn
        int unit_diagnostics
        int unit_error

    struct spral_lsmr_inform:
        int flag
        int itn
        int stat
        double normb
        double normAP
        double condAP
        double normr
        double normAPr
        double normy

    void spral_lsmr_default_options(spral_lsmr_options *options)

    int spral_lsmr_solve(int *action,
                         int m,
                         int n,
                         double u[],
                         double v[],
                         double y[],
                         void **keep,
                         const spral_lsmr_options *options,
                         spral_lsmr_inform *inform,
                         double *damp)

    int spral_lsmr_free(void **keep)


class ConvergenceTest(Enum):
    manual_simple = 1
    manual_norm = 2
    fond_saunders = 3


cdef class Options:
    cdef spral_lsmr_options options

    def __cinit__(self, **options):
        spral_lsmr_default_options(&self.options)

        for key, value in options.items():
            setattr(self, key, value)

    @property
    def atol(self):
        return self.options.atol

    @atol.setter
    def atol(self, value):
        self.options.atol = value

    @property
    def btol(self):
        return self.options.btol

    @btol.setter
    def btol(self, value):
        self.options.btol = value

    @property
    def conlim(self):
        return self.options.conlim

    @conlim.setter
    def conlim(self, value):
        self.options.conlim = value

    @property
    def ctest(self):
        return ConvergenceTest(self.options.ctest)

    @ctest.setter
    def ctest(self, value):
        self.options.ctest = ConvergenceTest[value].value

    @property
    def itnlim(self):
        return self.options.itnlim

    @itnlim.setter
    def itnlim(self, value):
        self.options.itnlim = value

    @property
    def itn_test(self):
        return self.options.itn_test

    @itn_test.setter
    def itn_test(self, value):
        self.options.itn_test = value

    @property
    def localSize(self):
        return self.options.localSize

    @localSize.setter
    def localSize(self, value):
        self.options.localSize = value

    @property
    def print_freq_head(self):
        return self.options.print_freq_head

    @print_freq_head.setter
    def print_freq_head(self, value):
        self.options.print_freq_head = value

    @property
    def print_freq_itn(self):
        return self.options.print_freq_itn

    @print_freq_itn.setter
    def print_freq_itn(self, value):
        self.options.print_freq_itn = value

    @property
    def unit_diagnostics(self):
        return self.options.unit_diagnostics

    @unit_diagnostics.setter
    def unit_diagnostics(self, value):
        self.options.unit_diagnostics = value

    @property
    def unit_error(self):
        return self.options.unit_error

    @unit_error.setter
    def unit_error(self, value):
        self.options.unit_error = value


class Status(Enum):
    zero_sol = 0
    compat_sol = 1
    suff_sol = 2
    cond_limit = 3
    res_suff_small = 4
    iter_limit = 7
    alloc_fail = 8
    dealloc_fail = 9
    neg_dimen = 10
    manual_abort = 11

    def error(self):
        error_codes = [Status.alloc_fail,
                       Status.dealloc_fail,
                       Status.neg_dimen]

        return self in error_codes

    @staticmethod
    def from_int(value):
        try:
            return Status(value)
        except ValueError:
            if value in [5, 6]:
                return Status.res_suff_small
            else:
                raise ValueError("Invalid status: {0}".format(value))

    def success(self):
        success_codes = [Status.zero_sol,
                         Status.compat_sol,
                         Status.suff_sol,
                         Status.cond_limit,
                         Status.res_suff_small]

        return self in success_codes


status_descriptions = {
    Status.zero_sol: "x = 0 is the exact solution. No iterations were performed.",
    Status.compat_sol: "The equations Ax = b are probably compatible. ||Ax - b|| is sufficiently small, given the values of `atol` and `btol`.",
    Status.suff_sol: "If lamb is zero then the system is probably not compatible. A least-squares solution has been obtained that is sufficiently accurate, given the value of `atol`. Otherwise, damped least-squares solution has been obtained that is sufficiently accurate, given the value of `atol`.",
    Status.cond_limit: "An estimate of cond(A) has exceeded `conlim`. The system appears to be ill-conditioned, or there could be an error in the products with A, A^T, P, or P^T.",
    Status.res_suff_small: "||APy - b|| is small enough for this machine.",
    Status.iter_limit: "Reached the iteration limit `itnlim` before the other stopping criteria were satisfied.",
    Status.alloc_fail: "Failure to allocate working space.",
    Status.dealloc_fail: "Failure to deallocate working space.",
    Status.neg_dimen: "One of the input dimensions is negative.",
    Status.manual_abort: "Computation was terminated by the user.",
}


class Result:
    def __init__(self, x, **props):
        self.x = x
        self.props = props

    def __getattr__(self, name):
        return self.props[name]


cpdef _solve(A: sp.sparse.linalg.LinearOperator,
             b: np.ndarray,
             P: Optional[sp.sparse.linalg.LinearOperator] = None,
             lamb: float = 0,
             options: Options = Options(),
             callback=None):
    cdef int action = 0
    cdef int m, n
    cdef double[::1] u_arr
    cdef double[::1] v_arr
    cdef double[::1] y_arr
    cdef void* keep = NULL
    cdef double damp = lamb
    cdef spral_lsmr_inform inform

    (m, n) = A.shape

    assert b.shape == (m,)

    if P is not None:
        assert P.shape == (n, n)

    u = np.copy(b)
    v = np.empty(shape=(n,))
    y = np.empty(shape=(n,))

    if options.ctest != ConvergenceTest.fond_saunders:
        if callback is None:
            raise ValueError("No callback provided for convergence test")
    else:
        if callback is not None:
            raise ValueError("Callback provided without convergence test")

    u_arr = u
    v_arr = v
    y_arr = y

    manual_abort = False

    try:
        while True:
            spral_lsmr_solve(&action,
                             m,
                             n,
                             &u_arr[0],
                             &v_arr[0],
                             &y_arr[0],
                             &keep,
                             &options.options,
                             &inform,
                             &damp)

            if action == 0:
                break
            elif action == 1:
                ATu = A.rmatvec(u)

                if P is not None:
                    v += P.rmatvec(ATu)
                else:
                    v += ATu

                continue
            elif action == 2:
                vv = v

                if P is not None:
                    vv = P.matvec(v)

                u += A.matvec(vv)
                continue
            elif action == 3:
                assert callback is not None

                itn = inform.itn

                if options.ctest == ConvergenceTest.manual_simple:
                    if not(callback(y, iteration=itn)):
                        manual_abort = True
                        break
                else:
                    assert options.ctest == ConvergenceTest.manual_norm

                    norms = {"condAP": inform.condAP,
                             "normAP": inform.normAP,
                             "normAPr": inform.normAPr,
                             "normr": inform.normr,
                             "normy": inform.normy}

                    if not(callback(y, iteration=itn, norms=norms)):
                        manual_abort = True
                        break

                continue
            else:
                raise ValueError("Invalid action: {0}".format(action))
    finally:
        spral_lsmr_free(&keep)

    if manual_abort:
        status = Status.manual_abort
    else:
        status = Status(inform.flag)

    if status.error():
        raise ValueError("Invalid status: {0}".format(status))

    return Result(y,
                  iterations=inform.itn,
                  status=status.name,
                  description=status_descriptions[status],
                  success=status.success())


def solve(A: sp.sparse.linalg.LinearOperator,
          b: np.ndarray,
          P: Optional[sp.sparse.linalg.LinearOperator] = None,
          lamb: float = 0,
          callback=None,
          **options):

    return _solve(A,
                  b,
                  P,
                  lamb,
                  Options(**options),
                  callback=callback)



#cython: language_level=3

from enum import Enum


class MatrixType(Enum):
    Unspecified = 0
    RealRect = 1
    ComplexRect = -1
    RealUnsymm = 2
    ComplexUnsymm = -2
    RealSymmPsdef = 3
    ComplexHermPsdef = -3
    RealSymmIndef = 4
    ComplexHermIndef = -4
    ComplexSymm = -5
    RealSkew = 6
    ComplexSkew = -6



import scipy as sp
import numpy as np

import pytest

from pyspral.ssids import analyze, solve, Job, PrintLevel, Ordering


@pytest.fixture
def indef_mat():
    mat = np.array([[ 2,  1,  0,  0,  0 ],
                    [ 1,  4,  1,  0,  1, ],
                    [ 0,  1,  3,  2,  0 ],
                    [ 0,  0,  2, -1,  0 ],
                    [ 0,  1,  0,  0,  2 ],],
                   dtype=float)

    return sp.sparse.csc_matrix(mat)


@pytest.fixture
def indef_rhs():
    return np.array([4.0, 17.0, 19.0, 2.0, 12.0])


def test_solve_indef(indef_mat, indef_rhs):
    mat = indef_mat

    symbolic_factor = analyze(mat, check=True)
    numeric_factor = symbolic_factor.factor(posdef=False)

    b = indef_rhs
    b_rhs = np.copy(b)

    sol = numeric_factor.solve(b_rhs, inplace=True)
    assert (sol == b_rhs).all()
    assert np.allclose(mat.dot(sol) - b, 0.)

    b_rhs = np.copy(b)
    sol = numeric_factor.solve(b_rhs, inplace=False)

    assert (b == b_rhs).all()
    assert np.allclose(mat.dot(sol) - b, 0.)


def test_scale(indef_mat, indef_rhs):
    mat = indef_mat
    scale = np.array([1., 2., 1., 2., 1.], dtype=float)

    symbolic_factor = analyze(mat, check=True)
    numeric_factor = symbolic_factor.factor(posdef=False, scale=scale)

    b = indef_rhs
    x = numeric_factor.solve(b, inplace=False)
    assert np.allclose(mat.dot(x) - b, 0.)


def test_options(indef_mat):
    mat = indef_mat
    analyze(mat,
            check=True,
            print_level=PrintLevel.BasicDiagnostic)


def test_order(indef_mat, indef_rhs):
    mat = indef_mat

    (_, n) = mat.shape

    order = np.arange(start=1, stop=(n + 1), dtype=np.intc)

    print(order)

    symbolic_factor = analyze(mat, check=True, order=order, ordering=Ordering.UserSupplied)
    numeric_factor = symbolic_factor.factor(posdef=False)

    b = indef_rhs
    x = numeric_factor.solve(b, inplace=False)
    assert np.allclose(mat.dot(x) - b, 0.)


def test_solve(indef_mat, indef_rhs):
    sol = solve(indef_mat, indef_rhs, posdef=False)
    assert np.allclose(indef_mat.dot(sol) - indef_rhs, 0.)


def test_solve_indef_multi(indef_mat):
    mat = indef_mat
    rhs = indef_rhs

    symbolic_factor = analyze(mat, check=True)
    numeric_factor = symbolic_factor.factor(posdef=False)

    (_, n) = mat.shape

    num_sols = 2

    rhs = np.eye(n)[:num_sols,:]
    sols_rhs = np.copy(rhs)

    sols = numeric_factor.solve(sols_rhs, inplace=True)

    for i in range(num_sols):
        assert np.allclose(mat.dot(sols[i, :]) - rhs[i, :], 0.)

    sols_rhs = np.copy(rhs)
    sols = numeric_factor.solve(sols_rhs, inplace=False)

    assert (rhs == sols_rhs).all()

    for i in range(num_sols):
        assert np.allclose(mat.dot(sols[i, :]) - rhs[i, :], 0.)

    rhs = np.eye(n + 1)[:num_sols, :n]

    sols_rhs = np.copy(rhs)
    sols = numeric_factor.solve(sols_rhs, inplace=False)

    for i in range(num_sols):
        assert np.allclose(mat.dot(sols[i, :]) - rhs[i, :], 0.)


def test_enquire(indef_mat):
    mat = indef_mat

    symbolic_factor = analyze(mat, check=True)
    numeric_factor = symbolic_factor.factor(posdef=False)

    (piv_order, d) = numeric_factor.enquire()

    (_, n) = mat.shape

    abs_piv_order = np.abs(piv_order)

    assert (abs_piv_order <= n).all()
    assert (abs_piv_order >= 1).all()


def test_inertia(indef_mat):
    mat = indef_mat

    eigvals = np.linalg.eigvalsh(mat.toarray())
    num_neg_evals = np.sum(eigvals < 0)

    symbolic_factor = analyze(mat, check=True)
    numeric_factor = symbolic_factor.factor(posdef=False)

    assert numeric_factor.inform.num_neg == num_neg_evals


def test_diag(indef_mat):
    mat = indef_mat

    symbolic_factor = analyze(mat, check=True)
    numeric_factor = symbolic_factor.factor(posdef=False)

    (_, n) = mat.shape

    rhs = np.eye(n)

    sol = numeric_factor.solve(rhs, job=Job.Dxb)

    assert np.allclose(sol, numeric_factor.D().toarray())



import scipy as sp
import numpy as np

import pytest

from pyspral.scaling import (scaling_auction,
                             scaling_hungarian,
                             scaling_equilib,
                             scaling,
                             ScalingError,
                             ScalingMethod)


@pytest.fixture
def symm_mat():
    mat = np.array([[ 2,  1,  0,  0,  0 ],
                    [ 1,  4,  1,  0,  1, ],
                    [ 0,  1,  3,  2,  0 ],
                    [ 0,  0,  2, -1,  0 ],
                    [ 0,  1,  0,  0,  2 ],],
                   dtype=float)

    return sp.sparse.csc_matrix(mat)


@pytest.fixture
def rect_mat():
    mat = np.array([[ 2,  1,  0,  0,  0 ],
                    [ 1,  4,  1,  0,  1, ],
                    [ 0,  1,  3,  2,  0 ]],
                   dtype=float)

    return sp.sparse.csc_matrix(mat)


@pytest.fixture
def singular_mat():
    mat = np.array([[ 2,  1,  0,  0,  0 ],
                    [ 1,  4,  1,  0,  1, ],
                    [ 0,  0,  0,  0,  0 ]],
                   dtype=float)

    return sp.sparse.csc_matrix(mat)


@pytest.fixture
def singular_symm_mat():
    mat = np.array([[ 2,  1,  0],
                    [ 1,  4,  0],
                    [ 0,  0,  0]],
                   dtype=float)

    return sp.sparse.csc_matrix(mat)


def check_dims_symm(symm_mat, result):
    (m, n) = symm_mat.shape
    assert result.scaling.shape == (n,)

    if result.matching is not None:
        assert result.matching.shape == (n,)


def check_dims_unsymm(unsymm_mat, result):
    (m, n) = unsymm_mat.shape
    assert result.row_scaling.shape == (m,)
    assert result.col_scaling.shape == (n,)

    if result.matching is not None:
        assert result.matching.shape == (n,)


def test_scaling_auction_symm(symm_mat):
    result = scaling_auction(symm_mat, symmetric=True)
    check_dims_symm(symm_mat, result)


def test_scaling_auction_options(symm_mat):
    result = scaling_auction(symm_mat, symmetric=True, max_iterations=10)
    check_dims_symm(symm_mat, result)


def test_scaling_auction_invalid_option(symm_mat):
    with pytest.raises(Exception):
        scaling_auction(symm_mat, symmetric=True, invalid_option=True)


def test_scaling_auction(rect_mat):
    result = scaling_auction(rect_mat)
    check_dims_unsymm(rect_mat, result)


def test_scaling_hungarian_symm(symm_mat):
    result = scaling_hungarian(symm_mat, symmetric=True)
    check_dims_symm(symm_mat, result)


def test_scaling_hungarian_options(symm_mat):
    result = scaling_hungarian(symm_mat, symmetric=True, scale_if_singular=True)
    check_dims_symm(symm_mat, result)


def test_scaling_hungarian_invalid_option(symm_mat):
    with pytest.raises(Exception):
        scaling_hungarian(symm_mat, symmetric=True, invalid_option=True)


def test_scaling_hungarian(rect_mat):
    result = scaling_hungarian(rect_mat)
    check_dims_unsymm(rect_mat, result)


@pytest.mark.skip(reason="Memory fault in spral (issue #200)")
def test_scaling_hungarian_singular(singular_mat, singular_symm_mat):
    with pytest.raises(ScalingError):
        scaling_hungarian(singular_mat, scale_if_singular=False)

    with pytest.raises(ScalingError):
        scaling_hungarian(singular_symm_mat, symmetric=True, scale_if_singular=False)

    with pytest.warns(RuntimeWarning):
        scaling_hungarian(singular_mat, scale_if_singular=True)

    with pytest.warns(RuntimeWarning):
        scaling_hungarian(singular_symm_mat, symmetric=True, scale_if_singular=True)


def test_scaling_equilib_symm(symm_mat):
    result = scaling_equilib(symm_mat, symmetric=True)
    check_dims_symm(symm_mat, result)


def test_scaling_equilib_options(symm_mat):
    result = scaling_equilib(symm_mat, symmetric=True, max_iterations=10, tol=1e-6)
    check_dims_symm(symm_mat, result)


def test_scaling_equilib_invalid_option(symm_mat):
    with pytest.raises(Exception):
        scaling_equilib(symm_mat, symmetric=True, invalid_option=True)


def test_scaling_equilib(rect_mat):
    result = scaling_equilib(rect_mat)
    check_dims_unsymm(rect_mat, result)


@pytest.mark.parametrize("scaling_method", list(ScalingMethod))
def test_scaling_symm(symm_mat, scaling_method):

    result = scaling(symm_mat, scaling_method, symmetric=True)
    check_dims_symm(symm_mat, result)


@pytest.mark.parametrize("scaling_method", list(ScalingMethod))
def test_scaling_unsymm(rect_mat, scaling_method):

    result = scaling(rect_mat, scaling_method, symmetric=False)
    check_dims_unsymm(rect_mat, result)



#cython: language_level=3str

from libc.stdint cimport int64_t
from enum import Enum, auto

cimport numpy as np
import numpy as np
import scipy as sp

from .types cimport spral_matrix_type
from .types import MatrixType


cdef extern from "spral.h":
    int SPRAL_RANDOM_INITIAL_SEED

    struct spral_rb_read_options:
       int array_base
       bint add_diagonal
       float extra_space
       int lwr_upr_full
       int values

    struct spral_rb_write_options:
       int array_base
       char val_format[21]

    void spral_rb_default_read_options(spral_rb_read_options *options)
    void spral_rb_default_write_options(spral_rb_write_options *options)

    int spral_rb_peek(const char *filename, int *m, int *n, int64_t *nelt, int64_t *nvar,
                      int64_t *nval, spral_matrix_type *matrix_type, char *type_code,
                      char *title, char *identifier)

    int spral_rb_read_ptr32(const char *filename, void **handle,
                            spral_matrix_type *matrix_type, int *m, int *n, int **ptr,
                            int **row, double **val, const spral_rb_read_options *options,
                            char *title, char *identifier, int *state)

    int spral_rb_write_ptr32(const char *filename,
                             spral_matrix_type matrix_type,
                             int m, int n, const int *ptr, const int *row, const double * val,
                             const spral_rb_write_options *options, const char *title,
                             const char *identifier)

    void spral_rb_free_handle(void **handle)


class LowerUpperValues(Enum):
    Lower = 1
    Upper = 2
    Full = 3


class ReadValues(Enum):
    Values = 0
    PatternOnly = 1
    ValuesOrRandSymm = 2
    ValuesOrRandSymmDiagDom = 3
    ValuesOrRandUnsymm = 4
    RandSymm = -2
    RandSymmDiagDom = -3
    RandUnsymm = -4


cdef class ReadOptions:
    cdef spral_rb_read_options options
    def __cinit__(self, **options):
        spral_rb_default_read_options(&self.options)

    def __init__(self, **options):
        for key, value in options.items():
            setattr(self, key, value)

    property array_base:
        def __get__(self): return self.options.array_base
        def __set__(self, bint value): self.options.array_base = value

    property add_diagonal:
        def __get__(self): return self.options.add_diagonal
        def __set__(self, bint value): self.options.add_diagonal = value

    property extra_space:
        def __get__(self): return self.options.extra_space
        def __set__(self, float value): self.options.extra_space = value

    property lwr_upr_full:
        def __get__(self): return LowerUpperValues(self.options.lwr_upr_full)
        def __set__(self, value): self.options.lwr_upr_full = value.value

    property values:
        def __get__(self): return ReadValues(self.options.values)
        def __set__(self, value): self.options.values = value.value


cdef class WriteOptions:
    cdef spral_rb_write_options options
    def __cinit__(self):
        spral_rb_default_write_options(&self.options)

    property array_base:
        def __get__(self): return self.options.array_base
        def __set__(self, value): self.options.array_base = value

    property val_format:
        def __get__(self): return self.options.val_format
        def __set__(self, value): self.options.val_format = value


class ExitCode(Enum):
    Success = 0
    FailedToOpen = -1
    InvalidFormat = -2
    IOError = -3
    InvalidDataType = -4
    ElementalFormat = -5
    InvalidMatrixType = -6
    InvalidOptionExtraSpace = -10
    InvalidOptionLwrUpr = -11
    InvalidOptionValues = -12
    AllocationFailure = -20


class ReadError(Exception):
    def __init__(self, filename, exit_code):
        self.filename = filename
        self.exit_code = exit_code

    def __str__(self):
        return "Error: reading {0}: {1}".format(self.filename,
                                                self.exit_code.name)


# class DataType(Enum):
#     Real = 'r'
#     Complex = 'c'
#     Integer = 'i'
#     Pattern = 'p'
#     AuxPattern = 'q'


# class MatrixType(Enum):
#     Symm = 's'
#     UnSymm = 'u'
#     Hermitian = 'h'
#     SkewSymm = 'z'
#     Rectangular = 'r'


# class MatrixEncoding(Enum):
#     CSC = 'a'
#     Elemental = 'e'


# class TypeCode:
#     def __init__(self,
#                  data_type,
#                  matrix_type,
#                  matrix_encoding):
#         self.data_type = data_type
#         self.matrix_type = matrix_type
#         self.matrix_encoding = matrix_encoding

#     def __str__(self):
#         d = self.data_type.name
#         m = self.matrix_type.name
#         e = self.matrix_encoding.name
#         return "TypeCode({0}, {1}, {2})".format(d, m, e)

#     @staticmethod
#     def from_string(name):
#         data_type = DataType(name[0])
#         matrix_type = MatrixType(name[1])
#         matrix_encoding = MatrixEncoding(name[2])
#         return TypeCode(data_type, matrix_type, matrix_encoding)


class MatrixInfo:
    def __init__(self,
                 m,
                 n,
                 nelt,
                 nvar,
                 nval,
                 matrix_type,
                 title,
                 identifier):
        self.m = m
        self.n = n
        self.shape = (m, n)
        self.nelt = nelt
        self.nvar = nvar
        self.nval = nval
        self.matrix_type = matrix_type
        self.title = title
        self.identifier = identifier


def _peek(str filename):
    cdef int m, n
    cdef int64_t nelt, nvar, nval
    cdef spral_matrix_type matrix_type
    cdef char type_code[4]
    cdef char title[73]
    cdef char identifier[9]
    cdef int return_code
    cdef bytes _filename = filename.encode('ascii')
    cdef const char* fname = _filename

    return_code = spral_rb_peek(fname,
                                &m,
                                &n,
                                &nelt,
                                &nvar,
                                &nval,
                                &matrix_type,
                                type_code,
                                title,
                                identifier)

    exit_code = ExitCode(return_code)

    if exit_code != ExitCode.Success:
        raise ReadError(filename, exit_code)

    return MatrixInfo(m,
                      n,
                      nelt,
                      nvar,
                      nval,
                      MatrixType(matrix_type),
                      (<bytes> title).decode('ascii'),
                      (<bytes> identifier).decode('ascii'))


def peek(filename):
    return _peek(str(filename))


cdef class Matrix:

    cdef dict __dict__

    def __init__(self, m, n, ptr, row, val, title, identifier, matrix_type):
        self.m = m
        self.n = n

        assert m >= 0
        assert n >= 0

        self.ptr = ptr
        self.row = row
        self.data = val

        assert self.ptr.shape == (self.n + 1,)

        assert self.nnz == ptr[-1]

        assert self.row.shape == (self.nnz,)
        assert self.data.shape == (self.nnz,)

        self.title = title
        self.identifier = identifier
        self.matrix_type = matrix_type

    def tocsc(self):
        ptr = np.copy(self.ptr)
        ptr[-1] = self.nnz
        return sp.sparse.csc_matrix((self.data, self.row, ptr),
                                    shape=self.shape)

    @property
    def nnz(self):
        (nnz,) = self.row.shape
        return nnz

    @property
    def shape(self):
        return (self.m, self.n)

    @staticmethod
    def fromcsc(matrix, title, identifier, matrix_type):

        if not sp.sparse.issparse(matrix):
            raise ValueError("Expected sparse matrix")

        m = matrix.shape[0]
        n = matrix.shape[1]

        indptr = np.copy(matrix.indptr)
        row = np.copy(matrix.indices)
        data = np.copy(matrix.data)

        return Matrix(m,
                      n,
                      indptr,
                      row,
                      data,
                      title,
                      identifier,
                      matrix_type)


def _read(str filename, ReadOptions options, int random_state):
    cdef int m, n
    cdef bytes _filename = filename.encode('ascii')
    cdef const char* fname = _filename
    cdef void* handle
    cdef spral_matrix_type matrix_type
    cdef int* ptr
    cdef int* row
    cdef double* val
    cdef char title[73]
    cdef char identifier[9]
    cdef int inform

    cdef int[:] row_view
    cdef int[:] ptr_view
    cdef double[:] data_view

    try:
        options.array_base = 0

        inform = spral_rb_read_ptr32(fname,
                                     &handle,
                                     &matrix_type,
                                     &m,
                                     &n,
                                     &ptr,
                                     &row,
                                     &val,
                                     &options.options,
                                     title,
                                     identifier,
                                     &random_state)

        code = ExitCode(inform)

        if code != ExitCode.Success:
            raise ReadError(filename, code)

        ptr_view = <int[:n + 1]> ptr

        nnz = ptr_view[-1]

        row_view = <int[:nnz]> row

        if val != NULL:
            data_view = <double[:nnz]> val
        else:
            data_view = None

        return Matrix(m,
                      n,
                      np.copy(ptr_view),
                      np.copy(row_view),
                      np.copy(data_view),
                      title=title.decode('ascii'),
                      identifier=identifier.decode('ascii'),
                      matrix_type=MatrixType(matrix_type))
    finally:
        spral_rb_free_handle(&handle)


def read(filename, random_state=None, **options):
    cdef ReadOptions read_options = ReadOptions(**options)
    cdef int rand_state = SPRAL_RANDOM_INITIAL_SEED

    if random_state is not None:
        rand_state = random_state

    return _read(str(filename), read_options, rand_state)


def _write(str filename, Matrix matrix, WriteOptions options):
    cdef int m, n
    cdef bytes _filename = filename.encode('ascii')
    cdef const char* fname = _filename
    cdef int[:] ptr = matrix.ptr
    cdef int[:] row = matrix.row
    cdef double[:] val = matrix.data
    cdef int inform
    cdef spral_matrix_type matrix_type_value = matrix.matrix_type.value

    m = matrix.m
    n = matrix.n

    options.array_base = 0

    if n < 0 or m < 0:
        raise ValueError("Negative dimensions")

    if matrix.ptr.shape != (n + 1,):
        raise ValueError("Invalid ptr shape")

    data_size = ptr[-1]

    if matrix.row.shape != (data_size,):
        raise ValueError("Invalid row shape")

    if matrix.data.shape != (data_size,):
        raise ValueError("Invalid data shape")

    inform = spral_rb_write_ptr32(fname,
                                  matrix_type_value,
                                  m,
                                  n,
                                  &ptr[0],
                                  &row[0],
                                  &val[0],
                                  &options.options,
                                  matrix.title.encode('ascii'),
                                  matrix.identifier.encode('ascii'))

    code = ExitCode(inform)

    if code != ExitCode.Success:
        raise ReadError(filename, code)



def write(filename, matrix, **options):
    cdef WriteOptions write_options = WriteOptions(**options)

    return _write(str(filename), matrix, write_options)



import scipy as sp
import numpy as np

from pyspral.ssmfe import solve_standard


class Laplacian(sp.sparse.linalg.LinearOperator):

    def __init__(self, m):
        self.m = m

        n = m * m

        super().__init__(dtype=float,
                         shape=(n, n))

    def _matvec(self, x):
        m = self.m

        y = np.empty_like(x)

        xv = x.reshape((m, m))
        yv = y.reshape((m, m))

        for i in range(m):
            for j in range(m):
                z = 4. * xv[i, j]

                if i > 0:
                    z -= xv[i - 1, j]
                if j > 0:
                    z -= xv[i, j - 1]
                if i + 1 < m:
                    z -= xv[i + 1, j]
                if j + 1 < m:
                    z -= xv[i, j + 1]

                yv[i, j] = z

        return y


class GaussSeidel(sp.sparse.linalg.LinearOperator):

    def __init__(self, m):
        self.m = m

        n = m * m

        super().__init__(dtype=float,
                         shape=(n, n))

    def _matvec(self, x):
        m = self.m

        y = x / 4.0

        yv = y.reshape((m, m))

        def zval(i, j):
            z = 0.0

            if i > 0:
                z += yv[i - 1, j]
            if j > 0:
                z += yv[i, j - 1]
            if i + 1 < m:
                z += yv[i + 1, j]
            if j + 1 < m:
                z += yv[i, j + 1]

            return z

        for i in range(m):
            for j in range(m):
                yv[i, j] = (yv[i, j] + zval(i, j) / 4.0) / 4.0

        for i in reversed(range(m)):
            for j in reversed(range(m)):
                yv[i, j] = (yv[i, j] + zval(i, j) / 4.0) / 4.0

        return y


def test_solve_standard():
    m = 20
    n = m * m

    # left = 5
    # mep = 2*left
    left = 1
    mep = 10

    A = Laplacian(m)

    result = solve_standard(A,
                            left,
                            mep)

    x = result.x
    lamb = result.lamb
    neig = x.shape[1]

    assert neig > 0
    assert lamb.shape == (neig,)

    for i in range(neig):
        r = A.dot(x[:, i]) - (lamb[i] * x[:, i])
        assert np.allclose(r, 0.)


def test_solve_precond():
    m = 20
    n = m * m

    # left = 5
    # mep = 2*left
    left = 1
    mep = 10

    A = Laplacian(m)
    P = GaussSeidel(m)

    result = solve_standard(A,
                            left,
                            mep,
                            P=P)

    x = result.x
    lamb = result.lamb
    neig = x.shape[1]

    assert neig > 0
    assert lamb.shape == (neig,)

    for i in range(neig):
        r = A.dot(x[:, i]) - (lamb[i] * x[:, i])
        assert np.allclose(r, 0.)



#cython: language_level=3str

cimport numpy as np

import numpy as np
import scipy as sp


def iter_spmatrix(matrix):
    if sp.sparse.isspmatrix_coo(matrix):
        for r, c, m in zip(matrix.row, matrix.col, matrix.data):
            yield r, c, m

    elif sp.sparse.isspmatrix_csc(matrix):
        for c in range(matrix.shape[1]):
            for ind in range(matrix.indptr[c], matrix.indptr[c+1]):
                yield matrix.indices[ind], c, matrix.data[ind]

    elif sp.sparse.isspmatrix_csr(matrix):
        for r in range(matrix.shape[0]):
            for ind in range(matrix.indptr[r], matrix.indptr[r+1]):
                yield r, matrix.indices[ind], matrix.data[ind]

    elif sp.sparse.isspmatrix_lil(matrix):
        for r in range(matrix.shape[0]):
            for c, d in zip(matrix.rows[r], matrix.data[r]):
                yield r, c, d
    else:
        raise NotImplementedError("Invalid matrix format")


cdef class _CSCMatrix:
    """
    CSC matrix as specified by SPRAL (one-indexed)
    """

    def __cinit__(self, mat, symmetric):
        (m, n) = mat.shape

        nnz = mat.getnnz()

        row_entries = [np.array([1], dtype=np.intc)] + [list() for _ in range(n)]
        data_entries = [list() for _ in range(n)]

        for r, c, v in iter_spmatrix(mat):
            if symmetric and (r < c):
                continue

            row_entries[c + 1].append(r + 1)
            data_entries[c].append(v)

        for c in range(n):
            row_entries[c + 1] = np.array(row_entries[c + 1],
                                          dtype=np.intc)

            data_entries[c] = np.array(data_entries[c],
                                       dtype=np.double)

            perm = np.argsort(row_entries[c + 1])
            row_entries[c + 1] = row_entries[c + 1][perm]
            data_entries[c] = data_entries[c][perm]

        self.rows = np.concatenate(row_entries[1:])
        self.cols = np.cumsum([row.size for row in row_entries], dtype=np.intc)
        self.data = np.concatenate(data_entries)
        self.shape = mat.shape



import numpy as np
import pytest

from pyspral.random_matrix import random_matrix_generate, MatrixType, Error


def test_random_matrix():
    m = 10
    n = 3
    nnz = 12

    mat = random_matrix_generate(m,
                                 n,
                                 nnz,
                                 MatrixType.Unspecified)

    assert mat.shape == (m, n)
    assert mat.nnz == nnz


def test_invalid_dim():
    n = 3
    nnz = 0

    with pytest.raises(Exception):
        random_matrix_generate(n,
                               n,
                               nnz,
                               MatrixType.Unspecified)


def test_insufficient_nnz():
    n = 3
    nnz = 2

    with pytest.raises(Error):
        random_matrix_generate(n,
                               n,
                               nnz,
                               MatrixType.Unspecified,
                               nonsingular=True)


def test_random_matrix_symm():
    n = 3
    nnz = 5

    mat = random_matrix_generate(n,
                                 n,
                                 nnz,
                                 MatrixType.RealSymmIndef)

    mat = mat.toarray()

    upper_indices = np.triu_indices(n, 1)

    upper_vals = mat[upper_indices]

    assert (upper_vals == 0.).all()


@pytest.mark.skip("Not implemented")
def test_random_spd():
    n = 3
    nnz = 3

    mat = random_matrix_generate(n,
                                 n,
                                 nnz,
                                 MatrixType.RealSymmPsdef,
                                 nonsingular=True)

    mat = mat.toarray()

    mat_sym = .5*(mat + mat.T)

    np.linalg.cholesky(mat_sym)
